#!/usr/bin/tclsh
#
# v_free.tcl, Auto Delete/Archive
set revision "Dec/20/2000"
 
#
# official homepage      : http://vware.hypermart.net
# author's email         : moopowah@yahoo.com
#
#
# :: Description ::
#     This is a nifty script to maintain the disk space on your unix
#     system.  It supports having multiple sections per partition, and
#     it can also archive instead of just deleting the oldest
#     directories.  you may have a huge raid with apps/games/ps2 on it
#     and you want to maintain the raid's disk space.  you simply list
#     those three dirs as srcDirs (in the configuration below) and the
#     overall oldest releases inside them is delete.  so if your oldest
#     two dirs among all three of those sections are a 2 month old app
#     and 1.8 month old game, then that is what will be deemed "old",
#     and depending on your configuration, they will be deleted/archived
#     to free some space.  you can tell it when a partition is deemed
#     "unclean" (i.e., it's low on disk space and needs a cleaning), and
#     while disk space is made, you specify how "clean" you want to make
#     it (i.e, how much disk space free defines a "clean" state).
#     generally speaking, you denote start/stop conditions.  i still use
#     it as of today (5/2/2002). :-)
#
# :: Requirements ::
#     1) unix utils: grep, du, df
#
# :: Installation ::
#     0) by default the script does not delete/archive, this is set so
#        you don't delete your data by a misconfiguration.  configure
#        this script (below) and just continue to step 1 for now. :-)
#     1) run the script manually to make sure it works as desired.  you
#        should see it attempting to delete/archive every dir in all the
#        srcDirs you specified if there is not enough free space.  set
#        the trigger free and stop free large enough so that each
#        device is evaluated (i.e., deemed to have too little space
#        available, for the sake of testing the script).
#     2) good, the script works.  there are nine (9) lines that you need
#        to uncomment before the script actually deletes/archives for
#        real.  search this file for "file delete", this will be 6 of
#        the lines.  you'll see a message telling you which lines to
#        uncomment.  then search for "file rename", this is the
#        remaining 3 lines to uncomment.  again, it tells you what to
#        do.
#     3) test the script manually one more time.  remember it will now
#        actually delete/archive, so if your stop condition is set too
#        high you're in for a disaster!
#     4) cron the script so it runs by itself, periodically:
#            */3 * * * *    /glftpd/bin/v_free.tcl 2>&1 >/dev/null 
#
# :: Todo ::
#     1) add an option to leave active symlinks to archive dir! :)  once
#        u del archive, u remove symlink.
#     2) replace "set currentFree [lindex [exec -- df -k | grep $device] 3]"
#        with a tcl procedure (to escape possible bug, and its cleaner?)
#
########################################################################


### Configuration ###

# 1 = archive the dirs
# 0 = deleted "old" dirs
set archive 0

# 1 = delete archive if need more free space to move in new releases
# 0 = don't free space, skip that particular move
set delArchive 1

# {<device> <triggerDiskFreeKB> <stopDiskFreeKB>
# <srcDir1>[:srcDir2:srcDirN] [archiveDir1[:archiveDir2:archiveDirN]]}
#                device: the mounted partition to check for disk usage
#     triggerDiskFreeKB: if the above device has this many or less
#                        KiloBytes of disk space free, then we will
#                        process this device (meaning, it needs to be
#                        cleaned!).  otherwise, the device has enough
#                        free space and we move on to the next one.
#                        stopDiskFreeKB: archive/delete (whatever you
#                        have set) the dirs in this device until this
#                        many KiloBytes of disk space is free.
#                srcDir: a list of comma-separated (:) dirs which
#                        contain the dirs which will be evaluated to
#                        determine what gets cleaned.  confusing?  it's
#                        not.  this exists in case you have both apps
#                        and games on the same partition and you want
#                        them to be equally "cleaned".  the oldest dirs
#                        out of all apps and games will be deleted; no
#                        segragation will be made (yay).
#            archiveDir: if you want to archive instead of delete, for
#                        "cleaning", then however many dirs you listed
#                        in srcDir, you need their respect archive dirs
#                        listed here.  once a dir is "old" in the
#                        srcDir, it is moved into its respective archive
#                        dir.  if you don't want to archive, then just
#                        set some junk value for this field (like
#                        {bla}).
set devices {
    {/dev/hdi1 300000 1000000 {/glftpd/site/movies/vcd} {bla}}
    {/dev/hdj2 300000 1000000 {/glftpd/site/movies/svcd} {bla}}
    {/dev/md0  300000 2000000 {/glftpd/site/games:/glftpd/site/tv:/glftpd/site/console/ps2_dvd} {bla}}
}

#  Example 1 (archive):
#  {/dev/ad5s1e 300000 1000000 {/glftpd/site/GAMES:/glftpd/site/LINUX-GAMES} {/glftpd/site/archive/cool_games:/glftpd/site/archive/linux/games}}

#  Example 2 (delete):
#  {/dev/ad7s1e 300000 1000000 {/glftpd/site/CONSOLE/DC:/glftpd/site/CONSOLE/PSX:/glftpd/site/CONSOLE/PS2} {bla}}
#  {/dev/ad4s1  300000 1000000 {/glftpd/site/MOVIES/DIVX:/glftpd/site/MOVIES/DVDS:/glftpd/site/MOVIES/XXX:/glftpd/site/MOVIES/VCDS} {bla}}



# "<dir to skip>" (wildcards are okay, case sensitive)
# These dirs will never be considered for archive/delete.
set ignoreDirs {
    "/glftpd/*lost+found"
    "/glftpd/\[*"
    "/glftpd/*NUKED*"
}

# If a release has this dir inside it, the release is ignored
# and not archived/deleted at all.  Example:
# /glftpd/site/apps/myApp/keep is a dir, then myApp is not
# archived or deleted even if it's deemed "old" by this script.
set ignoreRelease "keep"

### Configuration END ################################################################################








puts "advanced auto archive/delete script by vman (moopowah@yahoo.com); last revised $revision\n"

proc getOldDirs {srcDirs destDirs archive} {
        # Let's grab every release name from all srcDirs and put it in a list.
        set srcReleases ""
        for {set a 0} {$a < [llength $srcDirs]} {incr a} {
                if {![file isdirectory [lindex $srcDirs $a]]} {
                        puts "srcDir=\"[lindex $srcDirs $a]\" is not a directory; skipping tree."
                        continue
                }
                if {($archive == 1) && !([file isdirectory [lindex $destDirs $a]])} {
                        puts "destDir=\"[lindex $destDirs $a]\" is not a diretory; skipping tree."
                        continue
                }
                append srcReleases "[glob -nocomplain "[lindex $srcDirs $a]/*"] "
        }

        # Build another list with {dir mtime} and sort it based on mtime from least to greatest.
        set sortedSrcReleases ""
        foreach dir $srcReleases {
                lappend sortedSrcReleases "\{$dir\} [file mtime $dir]"
        }
        set sortedSrcReleases [lsort -increasing -integer -index 1 $sortedSrcReleases]
        return $sortedSrcReleases
}

proc enoughFreeSpace {dir amountNeeded} { global actualDevices mountPoints
        for {set a 0} {$a < [llength $mountPoints]} {incr a} {
                if {[string match "[lindex $mountPoints $a]*" $dir]} {
                        break
                }
        }
        set currentFree [expr [lindex [exec -- df -k | grep [lindex $actualDevices $a]] 3] / 1024]
        if {$currentFree < $amountNeeded} {
                upvar myDevice myDevice
                upvar myFreeSpace myFreeSpace
                set myDevice [lindex $actualDevices $a]
                set myFreeSpace $currentFree
                return 0
        }
        return 1
}





# Build a list with all mounted device names and a list if where devices
# are mounted (both will be respective to one another).
set temp ""
foreach diskMount [lrange [split [exec -- df -k] "\n"] 1 end] {
        lappend temp "[lindex $diskMount 0] [lindex $diskMount 5]"
}
set temp [lsort -decreasing -index 1 -dictionary $temp]
set actualDevices ""
set mountPoints ""
foreach myElement $temp {
        lappend actualDevices [lindex $myElement 0]
        lappend mountPoints [lindex $myElement 1]
}
unset temp

foreach {device triggerFree requiredFree srcDirs destDirs} [join $devices] {
        if {[lsearch $actualDevices $device] == -1} {
                # The device is not mounted, so let's skip it.
                continue
        }

        set srcDirs [split $srcDirs ":"]
        set destDirs [split $destDirs ":"]
        if {($archive == 1) && ([llength $srcDirs] != [llength $destDirs])} {
                puts "You don't have the same amount of srcDirs as you do destDirs; skipping device.\n"
                continue
        }

        set sortedSrcReleases [getOldDirs $srcDirs $destDirs $archive]

        # Cycle through the sorted list and move/delete the dirs until
	# you have enough free space, or exhausted the dirs for the
	# device.
        set currentFree [lindex [exec -- df -k | grep $device] 3]
        puts "${device}: currentFree=$currentFree, triggerFree=$triggerFree, requiredFree=$requiredFree; start..."
        if {$currentFree <= $triggerFree} {
                foreach {dir mtime} [join $sortedSrcReleases] {
                        if {$currentFree >= $requiredFree} {
                                break
                        }

                        # Check if our to-be-moved/deleted dir matches an ignoreDir or an ignoreRelease.
                        if {[file isdirectory "$dir/$ignoreRelease"]} {
                                # We want to save this puppy!
                                continue
                        }
                        set toSkip 0
                        foreach myIgnore $ignoreDirs {
                                if {[string match $myIgnore $dir]} {
                                        set toSkip 1
                                        break
                                }
                        }
                        if {$toSkip != 0} {
                                continue
                        }

                        set releaseSize [expr [lindex [exec -- du -s -k "$dir"] 0] / 1024]

                        # Either move dirs, or delete them.
                        if {$archive == 1} {
                                # Find out which srcDir the to-be-moved
				# dir resides in.  We will use its index
				# value for the destination tree (dir)
				# because srcDirs and destDirs are
				# respective.
                                set dirIndex ""
                                for {set a 0} {$a < [llength $srcDirs]} {incr a} {
                                        if {[string match "[lindex $srcDirs $a]/*" $dir]} {
                                                set dirIndex $a
                                                break
                                        }
                                }
                                set myDest [lindex $destDirs $dirIndex]

                                # Verify the destination has enough free
				# space (add an extra 20MB as a safety
				# buffer).
                                if {![enoughFreeSpace $myDest [expr $releaseSize + 20]]} {
                                        if {$delArchive == 1} {
                                                set oldDirs [getOldDirs $myDest "" 0]
                                                for {set a 0} {$myFreeSpace < [expr $releaseSize + 20]} {incr a} {
                                                        set myReleaseSize [expr [lindex [exec -- du -s -k "[lindex [lindex $oldDirs $a] 0]"] 0] / 1024]
                                                        puts -nonewline "${device}: deleting *archive* ${myReleaseSize}MB \"[lindex [lindex $oldDirs $a] 0]\" (in order to move \"$dir\")..."; flush stdout

                                                        # IF YOU'RE READY TO USE THIS SCRIPT, UNCOMMENT THE NEXT THREE LINES.
                                                        #if {[catch {file delete -force -- [lindex [lindex $oldDirs $a] 0]} err]} {
                                                        #       puts $err
                                                        #}

                                                        puts "done."
                                                        set myFreeSpace [expr [lindex [exec -- df -k | grep $myDevice] 3] / 1024]
                                                }
                                        } else {
                                                # If device is full and we can't delete, skip to another release.
                                                puts "${device}: can't move ${releaseSize}MB \"$dir\" -> \"$myDest/\" because destination is full!"
                                                continue
                                        }
                                }

                                # Move the release to its destination.
                                puts -nonewline "${device}: moving ${releaseSize}MB \"$dir\" -> \"$myDest/\"..."; flush stdout
                                # IF YOU'RE READY TO USE THIS SCRIPT, UNCOMMENT THE NEXT THREE LINES.
                                #if {[catch {file rename -force -- $dir "$myDest/"} err]} {
                                #       puts $err
                                #}
                        } else {
                                # Let's delete!
                                puts -nonewline "${device}: deleting ${releaseSize}MB \"$dir\"..."; flush stdout

                                # IF YOU'RE READY TO USE THIS SCRIPT, UNCOMMENT THE NEXT THREE LINES.
                                #if {[catch {file delete -force -- $dir} err]} {
                                #       puts $err
                                #}
                        }
                        puts "done."

                        # Bug alert: let's hope the device name is not
			# used as a mount point for another device!
                        set currentFree [lindex [exec -- df -k | grep $device] 3]
                }
        }
        puts "${device}: currentFree=$currentFree, triggerFree=$triggerFree, requiredFree=$requiredFree; done with device!\n"
}

# EOF. :-)
